/*
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.



	This is a port of madfuload (http://usb-midi-fw.sourceforge.net/)
	using libusb (http://libusb.wiki.sourceforge.net/)
	
	madfuloader was written by Joakim L. Gilje
	madfuload was written by Clemens Ladisch
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <syslog.h>
#include <getopt.h>

#include "usb.h"

#define USB_DT_INTERFACE	0x04
#define USB_CLASS_APP_SPEC	0xfe

/* symbols and types from the DFU standard */
#define USB_SUBCLASS_DFU	0x01
#define USB_DT_DFU_FUNCTIONAL	0x21

#define DFU_CAN_DNLOAD			0x01
#define DFU_CAN_UPLOAD			0x02
#define DFU_MANIFESTATION_TOLERANT	0x04

#define DFU_DETACH      0
#define DFU_DNLOAD      1
#define DFU_UPLOAD      2
#define DFU_GETSTATUS   3
#define DFU_CLRSTATUS   4
#define DFU_GETSTATE    5
#define DFU_ABORT       6

enum {
	WAIT_STANDARD,
	WAIT_IGNORE,
	WAIT_BYTE3
};

static int to_logger;
static int wait_type;
static char *command;
static char *firmware;
static int firmware_length;
static int transfer_size = 64;
static int logging;
static int vendorID = -1;
static int productID = -1;

int interface;

struct usb_device *usbdevice;
struct usb_dev_handle *usbhandle;
struct usb_interface_descriptor *usbinterface;

static void print(int level, const char *format, va_list ap)
{
	if (to_logger) {
		vsyslog(LOG_DAEMON | level, format, ap);
	} else {
		vfprintf(stderr, format, ap);
		putc('\n', stderr);
	}
}

static void fatal(const char *format, ...)
{
	va_list ap;

	va_start(ap, format);
	print(LOG_ERR, format, ap);
	va_end(ap);
	exit(1);
}

static void warning(const char *format, ...)
{
	va_list ap;

	va_start(ap, format);
	print(LOG_WARNING, format, ap);
	va_end(ap);
}

static void logmsg(const char *format, ...)
{
	va_list ap;

	if (!logging)
		return;
	va_start(ap, format);
	print(LOG_DEBUG, format, ap);
	va_end(ap);
}

static void load_firmware(char *filename)
{
	int fd = open(filename, O_RDONLY);
	if (fd == -1)
		fatal("cannot open %s: %s", filename, strerror(errno));
	firmware_length = lseek(fd, 0, SEEK_END);
	if (firmware_length < 0)
		fatal("cannot determine length of %s: %s", filename, strerror(errno));
	if (firmware_length == 0)
		fatal("firmware file %s is empty", filename);
	lseek(fd, 0, SEEK_SET);
	firmware = malloc(firmware_length);
	if (!firmware)
		fatal("out of memory");
	if (read(fd, firmware, firmware_length) != firmware_length)
		fatal("cannot read from %s: %s", filename, strerror(errno));
	close(fd);
	logmsg("%s: %d bytes read successfully", filename, firmware_length);
}

static void findInterface(void)
{
    int i, j, k;
	int interface_found = 0;
        
    printf("usbdevice->descriptor.bDescriptorType: %x\n", usbdevice->descriptor.bDescriptorType);
	for (i = 0; i < usbdevice->descriptor.bNumConfigurations; i++) {
        for (j = 0; j < usbdevice->config[i].bNumInterfaces; j++) {
            for (k = 0; k < usbdevice->config[i].interface[j].num_altsetting; k++) {
                struct usb_interface_descriptor *interface = &usbdevice->config[i].interface[j].altsetting[k];
                if ((interface->bInterfaceClass == USB_CLASS_APP_SPEC) &&
                    (interface->bInterfaceSubClass == USB_SUBCLASS_DFU)) {
                    interface_found = 1;
                    usbinterface = interface;
                    logmsg("DFU interface is %d", interface->bInterfaceNumber);
                } else {
                    interface_found = 0;
                }
            }
        }
    }
    return;
/*
		} else if (desc->bDescriptorType == USB_DT_DFU_FUNCTIONAL &&
			   desc->bLength >= 7 &&
			   interface_found) {
			unsigned char d[5];
			logmsg("DFU descriptor found");

			transfer_size = d[3] | (d[4] << 8);
			if (transfer_size < 1)
				fatal("invalid transfer size %d", transfer_size);
			logmsg("transfer size is %d", transfer_size);
		}
*/
}

static void dfu_dnload(void *buf, int length, int block)
{
	int err;
	
    err = usb_control_msg(usbhandle,
                   USB_ENDPOINT_OUT | USB_TYPE_CLASS | USB_RECIP_INTERFACE,
                   DFU_DNLOAD,
                   block, 
                   usbinterface->bInterfaceNumber,
                   buf,
                   length,
                   500);

	if (err < 0)
		fatal("downloading block %d failed", block);
}

static void dfu_status_wait()
{
	int wait, err;
	char status[6];
    
    err = usb_control_msg(usbhandle,
                          USB_ENDPOINT_IN | USB_TYPE_CLASS | USB_RECIP_INTERFACE,
                          DFU_GETSTATUS,
                          0,
                          usbinterface->bInterfaceNumber,
                          status,
                          6,
                          500);

	if (err < 0)
		fatal("cannot get device status");
    
    switch (wait_type) {
        case WAIT_BYTE3:
        {
            struct timespec tv;
            wait = status[3];
            logmsg("waiting %d ms", wait);
            tv.tv_sec = wait / 1000;
            tv.tv_nsec = (wait % 1000) * 1000000;
            if (nanosleep(&tv, NULL) == -1)
                fatal("sleep failed: %s", strerror(errno));
            
            break;
        }
        default:
            break;
    }
}

static void download_firmware()
{
	unsigned short block = 0;
	int pos, length;
    
    usbhandle = usb_open(usbdevice);
	if (usbhandle == NULL)
		fatal("failed to open device (%s)", usb_strerror());

	findInterface();
	if (usb_claim_interface(usbhandle, usbinterface->bInterfaceNumber) < 0)
		fatal("cannot claim interface %d: (%d) %s", usbinterface->bInterfaceNumber, errno, strerror(errno));

	for (pos = 0; pos < firmware_length; pos += transfer_size) {
		if (firmware_length - pos > transfer_size)
			length = transfer_size;
		else
			length = firmware_length - pos;
		dfu_dnload(firmware + pos, length, block);
		++block;

		dfu_status_wait();
	}
	dfu_dnload(firmware, 0, block);
	dfu_status_wait();

	if (usb_reset(usbhandle) == -1)
		// might fail because the device has already reset itself ...
		warning("cannot reset device: (%d) %s", errno, strerror(errno));
	
    usb_close(usbhandle);
}

struct usb_device* scanBus() {
    struct usb_bus *bus;
    
    usb_find_busses();
    usb_find_devices();
    
    for (bus = usb_busses; bus; bus = bus->next) {
        struct usb_device *dev;
            
        for (dev = bus->devices; dev; dev = dev->next) {
            printf("idProduct: %x, idVendor: %x, %d\n",
                                    dev->descriptor.idProduct,
                                    dev->descriptor.idVendor,
                                    dev->descriptor.bDeviceClass);
            if ((dev->descriptor.idProduct == productID) &&
                (dev->descriptor.idVendor == vendorID)) {
                return dev;
            }
        }
    }
    
    return NULL;
}

static void help() {
	fprintf(stderr,
		"Usage: %s options...\n"
		"\n"
		"-h, --help           help\n"
		"-V, --version        print version\n"
		"-f, --firmware=path  file containing the firmware to download\n"
		"-d, --vendor=id      VendorID of device\n"
        "-p, --product=id     ProductID of device\n"
		"-v, --verbose        output debugging messages\n"
		"-3, --waitbyte3      use the third byte of the bwPollTimeout field\n",
		command);
}

static void version() {
	fprintf(stderr, "%s version 0.1\n", command);
}

int main (int argc, char* const *argv) {
	static char short_options[] = "hVf:d:p:v3";
	static struct option long_options[] = {
		{"help",      0, NULL, 'h'},
		{"version",   0, NULL, 'V'},
		{"firmware",  1, NULL, 'f'},
		{"vendor",    1, NULL, 'd'},
		{"product",   1, NULL, 'p'},
		{"verbose",   0, NULL, 'v'},
		{"waitbyte3", 0, NULL, '3'},
		{NULL}
	};
    
	char *firmware = NULL;
	int c;

	command = argv[0];
    wait_type = WAIT_IGNORE;
	while ((c = getopt_long(argc, argv, short_options,
				long_options, NULL)) != -1) {
		switch (c) {
		case 'h':
			help();
			return 0;
		case 'V':
			version();
			return 0;
		case 'f':
			firmware = optarg;
			break;
		case 'd':
			vendorID = strtol(optarg, NULL, 16);
			break;
		case 'p':
			productID = strtol(optarg, NULL, 16);
			break;
		case 'v':
			logging = 1;
			break;
		case 'l':
			to_logger = 1;
			break;
		case '3':
			wait_type = WAIT_BYTE3;
			break;
		default:
			fprintf(stderr, "Try %s --help for more information.\n", command);
			return 1;
		}
	}
	if (!firmware) {
		fputs("Please specify a firmware file.\n", stderr);
		return 1;
	}
	if ((vendorID < 0) || (productID < 0))
		fatal("specify both a vendorID and productID");

    usb_init();
    load_firmware(firmware);
    usbdevice = scanBus();
    if (! usbdevice) {
        fatal("no device found");
    }

    download_firmware();
    
    return 0;
}
